import sys
from pathlib import Path

FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))
    
from defaults.featuring_defaults import *
from defaults.train_defaults import BATCH_SIZE
from .conll2003 import load_data
from .dataset import ConllDataset, PredictDataset
from .featuring import Featuring
from .tokenizer import tokenize

import numpy as np
import torch
from torch.nn.functional import pad
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader

def _custom_collate_fn(batch):
    """
    Function for sorting and padding each batch data for packed_pad_sequence()
    in train, validation, test phases.
    """
    new_batch = {"features": {}}
    keys = batch[0]["features"].keys()
    sorted_batch = sorted(
        [sample for sample in batch],
        key=lambda x: x["features"]["global_vectors"].shape[0],
        reverse=True,
    )
    for key in keys:

        if key == "casing_vectors":
            new_batch["features"][key] = pad_sequence(
                [
                    torch.Tensor(np.array(sample["features"][key]))
                    for sample in sorted_batch
                ],
                batch_first=True,
            )

        elif key == "global_vectors":
            new_batch["features"][key] = pad_sequence(
                [
                    torch.Tensor(np.array(sample["features"][key]))
                    for sample in sorted_batch
                ],
                padding_value=1,
                batch_first=True,
            )
            
            new_batch["features"]["seq_lengths"] = [
                sample["features"][key].shape[0] for sample in sorted_batch
            ]

        elif key == "char_level_vectors":
            max_len_word = max([len(word) for word in sorted_batch[0]["features"][key]])
            max_len_sequence = sorted_batch[0]["features"]["global_vectors"].shape[0]

            pad_words = lambda words: pad_sequence(
                [torch.Tensor(word) for word in words], batch_first=True
            )
            padded_samples = [
                pad_words(sample["features"][key]) for sample in sorted_batch
            ]

            pad_tensor = lambda tensor: pad(
                tensor,
                (
                    0,
                    max_len_word - tensor.shape[1],
                    0,
                    max_len_sequence - tensor.shape[0],
                ),
                "constant",
                0,
            )
            new_batch["features"][key] = torch.stack(
                [pad_tensor(sample) for sample in padded_samples]
            )
            new_batch["features"]["seq_lengths"] = [
                sample.shape[0] for sample in padded_samples
            ]

    if sorted_batch[0].get("tags", -1) != -1:
        new_batch["tags"] = pad_sequence(
            [torch.Tensor(sample["tags"]) for sample in sorted_batch], batch_first=True
        )
    return new_batch


def load_transform_data(
    phase,
    use_char_level=USE_CHAR_LEVEL_EMBEDDING,
    use_casing=USE_CASING_FEATURE,
    batch_size=BATCH_SIZE,
):
    """
    Load raw data then transform it into features.
    """
    raw_dataset = load_data()[phase]

    datafeaturing = Featuring(use_char_level=use_char_level, use_casing=use_casing)

    dataset = ConllDataset(dataset=raw_dataset, preprocessing=datafeaturing)

    use_shuffle = phase == "train"

    dataloader = DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        shuffle=use_shuffle,
        collate_fn=_custom_collate_fn,
    )

    return dataloader


def load_from_array(
    sequences,
    use_char_level=USE_CHAR_LEVEL_EMBEDDING,
    use_casing=USE_CASING_FEATURE,
    batch_size=1,
):
    """
    Transform sequences to features.
    """
    sequences = tokenize(sequences)
    datafeaturing = Featuring(use_char_level=use_char_level, use_casing=use_casing)

    dataset = PredictDataset(dataset=sequences, preprocessing=datafeaturing)

    use_shuffle = False

    dataloader = DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        shuffle=use_shuffle,
        collate_fn=_custom_collate_fn,
    )
    return dataloader



GLOVE_DIM = 100
USE_CHAR_LEVEL_EMBEDDING = True
USE_CASING_FEATURE = True
PRETRAIN_PATH = "pretrain/glove.6B.100d.txt"
ALL_CHARACTERS = " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_`abcdefghijklmnopqrstuvwxyz{|}~"



from torchtext.data.utils import get_tokenizer

def tokenize(corpus):
    # tokenize sequence
    tokenizer = get_tokenizer("spacy")
    tokens_sequences = []
    for sentence in corpus:
        tokens_sequences.append(tokenizer(sentence))

    return tokens_sequences



import sys
from pathlib import Path
FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))

import argparse
import val as validate
from defaults.model_defaults import MODEL_MODE
from defaults.train_defaults import *
from helpers.callbacks import CustomCallback
from helpers.logger import log_configuration
from model import NERModel
from model_configs.configs import get_configs
from prepare.data import load_transform_data

import torch
from torch.nn.utils import clip_grad_norm_
from torch.optim import Adam

def train(model_configs, train_config):
    """
    Create the model base on model config dict then train
    the model with settings in train config dict.
    """
    device_name = "cpu"
    device = torch.device(device_name)
    if train_config["device_mode"] == 1:
        device_name = "cuda" if torch.cuda.is_available() else "cpu"
        device = torch.device(device_name)

    model = NERModel(
        model_configs["lstm"],
        model_configs["char_level_embedding"],
        model_configs["casing_embedding"],
        model_configs["crf"],
        device,
    ).to(device)
    optimizer = Adam(model.parameters(), lr=train_config["learning_rate"])

    callback = CustomCallback(
        model_name=model_configs["model_name"],
        optimizer=optimizer,
        epochs=train_config["epochs"],
        factor=train_config["factor"],
        decay_patience=train_config["decay_patience"],
        mode=train_config["mode"],
        es_patience=train_config["es_patience"],
    )

    train_dataloader = load_transform_data(
        "train",
        use_char_level=model_configs["use_char_level_embedding"],
        use_casing=model_configs["use_casing_feature"],
    )

    validation_dataloader = load_transform_data(
        "validation",
        use_char_level=model_configs["use_char_level_embedding"],
        use_casing=model_configs["use_casing_feature"],
    )
    num_steps = len(train_dataloader)

    log_configuration("train", model_configs["model_name"], train_config, device_name)
    log_configuration(
        "validation", model_configs["model_name"], train_config, device_name
    )
    for epoch in range(train_config["epochs"]):
        epoch_loss = 0
        model.train()
        callback.reset_state()
        for i, batch in enumerate(train_dataloader):
            optimizer.zero_grad()
            loss = model.fit_and_compute_loss(batch)
            epoch_loss += loss
            loss.backward()
            clip_grad_norm_(model.parameters(), train_config["gradient_clipping"])
            optimizer.step()

        epoch_loss = epoch_loss / num_steps

        model_state_dict = model.state_dict()
        f1 = validate.train_validate(
            model_configs, train_config, model_state_dict, validation_dataloader
        )
        learning_rate = optimizer.param_groups[0]["lr"]

        stop_value = callback.step(
            epoch, epoch_loss, f1, learning_rate, model_state_dict
        )

        if stop_value == 1:
            break

def parse_opt():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-mode", type=int, default=MODEL_MODE, help="mode of model"
    )
    parser.add_argument("--epochs", type=int, default=EPOCHS, help="training epochs")
    parser.add_argument("--batch-size", type=int, default=BATCH_SIZE, help="batch size")
    parser.add_argument(
        "--learning-rate", type=float, default=LEARNING_RATE, help="learning rate"
    )
    parser.add_argument(
        "--device-mode", type=int, default=USE_GPU, help="0: cpu - 1: cuda"
    )
    return parser.parse_args()

def main(opt):
    # Create model config and train config from parser and defaults
    model_config = get_configs(opt.model_mode)
    train_config = {}
    train_config["gradient_clipping"] = GRADIENT_CLIPPING
    train_config["factor"] = FACTOR
    train_config["decay_patience"] = DECAY_PATIENCE
    train_config["mode"] = MODE
    train_config["es_patience"] = ES_PATIENCE
    train_config["epochs"] = opt.epochs
    train_config["batch_size"] = opt.batch_size
    train_config["learning_rate"] = opt.learning_rate
    train_config["device_mode"] = opt.device_mode

    train(model_config, train_config)

if __name__ == "__main__":
    opt = parse_opt()
    main(opt)



MODEL_MODE = 5

CHAR_EMBEDDING_NUM_VOCABS = 100
CHAR_EMBEDDING_VECTORIZE_DIM = 50

CNN_EMBEDDING_IN_CHANNELS = 1
CNN_EMBEDDING_OUT_CHANNELS = 50
CNN_EMBEDDING_KERNEL_SIZE = (3, 50)
CNN_EMBEDDING_PADDING_SIZE = (1, 0)

LSTM_EMBEDDING_HIDDEN_SIZE = 25
LSTM_EMBEDDING_NUM_LAYERS = 1
LSTM_EMBEDDING_BIDIRECTIONAL = True
CASING_NUM_VOCABS = 8
CASING_VECTORIZE_DIM = 50
LSTM_HIDDEN_SIZE = 275
P_DROPOUT = 0.5
CRF_TARGET_SIZE = 11



import sys
from pathlib import Path

FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))

import argparse

import torch
from defaults.model_defaults import MODEL_MODE
from defaults.tag_dict import *
from defaults.train_defaults import EPOCHS, USE_GPU
from helpers.logger import log_configuration, log_test_result
from model import NERModel
from model_configs.configs import get_configs
from prepare.data import load_transform_data
from seqeval.metrics import (classification_report, f1_score, precision_score,
                             recall_score)


def train_validate(model_configs, train_config, state_dict, dataloader):
    """
    Function for help train phase to validate epoch's model on validation dataset:
    :model_configs, train_config, state_dict: given by train phase
    :dataloader: Dataloader instance of validation dataset
    """
    device_name = "cpu"
    device = torch.device(device_name)
    if train_config["device_mode"] == 1:
        device_name = "cuda" if torch.cuda.is_available() else "cpu"
        device = torch.device(device_name)

    model = NERModel(
        model_configs["lstm"],
        model_configs["char_level_embedding"],
        model_configs["casing_embedding"],
        model_configs["crf"],
        device,
    ).to(device)

    model.load_state_dict(state_dict)
    model.eval()

    with torch.no_grad():
        all_predicts = []
        true_tags = []
        for i, batch in enumerate(dataloader):
            best_paths = model.forward(batch)
            seq_lengths = batch["features"]["seq_lengths"]
            idxs_to_tags = lambda idxs: [IDX_TO_TAG[idx] for idx in idxs]
            for j in range(len(batch)):

                sample_predicts = best_paths[j][: seq_lengths[j]]

                sample_true_tags = batch["tags"][j][: seq_lengths[j]].long().tolist()

                all_predicts.append(idxs_to_tags(sample_predicts))
                true_tags.append(idxs_to_tags(sample_true_tags))

        f1 = f1_score(true_tags, all_predicts)
        return f1


def run(model_configs, weights_path, device_mode):
    """
    Validate chosen model with weights saved in weights_path.
    """
    device_name = "cpu"
    device = torch.device(device_name)
    if device_mode == 1:
        device_name = "cuda" if torch.cuda.is_available() else "cpu"
        device = torch.device(device_name)

    if weights_path == "":
        weights_path = ROOT / "models" / model_configs["model_name"] / "best.w"
    else:
        weights_path = ROOT / "models" / model_configs["model_name"] / weights_path
    model = NERModel(
        model_configs["lstm"],
        model_configs["char_level_embedding"],
        model_configs["casing_embedding"],
        model_configs["crf"],
        device,
    ).to(device)

    model.load_state_dict(torch.load(weights_path))
    model.eval()

    dataloader = load_transform_data(
        "test",
        use_char_level=model_configs["use_char_level_embedding"],
        use_casing=model_configs["use_casing_feature"],
    )

    log_configuration("test", model_configs["model_name"], None, device_name)

    with torch.no_grad():
        all_predicts = []
        true_tags = []
        for i, batch in enumerate(dataloader):
            best_paths = model.forward(batch)
            seq_lengths = batch["features"]["seq_lengths"]
            idxs_to_tags = lambda idxs: [IDX_TO_TAG[idx] for idx in idxs]
            for j in range(len(batch)):

                sample_predicts = best_paths[j][: seq_lengths[j]]
                sample_true_tags = batch["tags"][j][: seq_lengths[j]].long().tolist()

                all_predicts.append(idxs_to_tags(sample_predicts))
                true_tags.append(idxs_to_tags(sample_true_tags))

        f1 = round(f1_score(true_tags, all_predicts) * 100, 2)
        precision = round(precision_score(true_tags, all_predicts) * 100, 2)
        recall = round(recall_score(true_tags, all_predicts) * 100, 2)
        
        log_test_result(f1, precision, recall, classification_report(true_tags, all_predicts))



def parse_opt():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-mode", type=int, default=MODEL_MODE, help="mode of model"
    )
    parser.add_argument("--weights", type=str, default="", help="weights file name")
    parser.add_argument(
        "--device-mode", type=int, default=USE_GPU, help="0: cpu - 1: cuda"
    )
    return parser.parse_args()


def main(opt):
    model_configs = get_configs(opt.model_mode)
    weights_path = opt.weights
    device_mode = opt.device_mode
    run(model_configs, weights_path, device_mode)


if __name__ == "__main__":
    opt = parse_opt()
    main(opt)



import sys
from pathlib import Path

FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))
    
from defaults.featuring_defaults import GLOVE_DIM
from defaults.model_defaults import *
from helpers.config import *

import json

def get_configs(mode):

    """
    Read Supported Models in configs.json then create a dict
    to save chosen model settings.
    """
    lstm_input_size = GLOVE_DIM
    f = open(ROOT / "model_configs/configs.json")
    model_dict = json.load(f)
    model_config = list(model_dict.values())[mode]

    model_config["model_name"] = list(model_dict.keys())[mode]
    model_config["char_level_embedding"] = None
    model_config["casing_embedding"] = None
    model_config["crf"] = None
    if model_config["use_char_level_embedding"]:

        if model_config["use_char_level_embedding"] == "CNN":
            model_config["char_level_embedding"] = create_cnn_embedding_config(
                CHAR_EMBEDDING_NUM_VOCABS,
                CHAR_EMBEDDING_VECTORIZE_DIM,
                CNN_EMBEDDING_IN_CHANNELS,
                CNN_EMBEDDING_OUT_CHANNELS,
                CNN_EMBEDDING_KERNEL_SIZE,
                CNN_EMBEDDING_PADDING_SIZE,
            )
            lstm_input_size += CNN_EMBEDDING_OUT_CHANNELS

        elif model_config["use_char_level_embedding"] == "LSTM":
            model_config["char_level_embedding"] = create_lstm_embedding_config(
                CHAR_EMBEDDING_NUM_VOCABS,
                CHAR_EMBEDDING_VECTORIZE_DIM,
                LSTM_EMBEDDING_HIDDEN_SIZE,
                LSTM_EMBEDDING_NUM_LAYERS,
                LSTM_EMBEDDING_BIDIRECTIONAL,
            )
            lstm_input_size += 2 * LSTM_EMBEDDING_HIDDEN_SIZE

    if model_config["use_casing_feature"]:
        model_config["casing_embedding"] = create_casing_config(
            CASING_NUM_VOCABS, CASING_VECTORIZE_DIM
        )
        lstm_input_size += CASING_VECTORIZE_DIM

    if model_config["use_crf"]:
        model_config["crf"] = create_crf_config(CRF_TARGET_SIZE)

    model_config["lstm"] = create_lstm_config(
        lstm_input_size, LSTM_HIDDEN_SIZE, P_DROPOUT
    )

    return model_config



from torch.utils.data import Dataset

class ConllDataset(Dataset):
    def __init__(self, dataset, preprocessing):
        self.dataset = dataset
        self.preprocessing = preprocessing

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        tokens = self.dataset[index]["tokens"]
        tags = self.dataset[index]["ner_tags"]
        converted_sample = self.preprocessing.preprocess(tokens, tags)
        return converted_sample

class PredictDataset(Dataset):
    # Dataset for predict sequences.
    def __init__(self, dataset, preprocessing):
        self.dataset = dataset
        self.preprocessing = preprocessing

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        tokens = self.dataset[index]
        converted_sample = self.preprocessing.preprocess(tokens, None)
        return converted_sample



""" 
  Functions for initialize parameters of each layer in model,
  then save into dict object.
"""
def create_cnn_embedding_config(
    num_vocabs, vectorize_dim, in_channels, out_channels, kernel_size, padding_size
):
    config = {"mode": "CNN"}
    config["num_vocabs"] = num_vocabs
    config["vectorize_dim"] = vectorize_dim
    config["in_channels"] = in_channels
    config["out_channels"] = out_channels
    config["kernel_size"] = kernel_size
    config["padding_size"] = padding_size

    return config


def create_lstm_embedding_config(
    num_vocabs, vectorize_dim, hidden_size, num_layers, bidirectional
):
    config = {"mode": "LSTM"}
    config["num_vocabs"] = num_vocabs
    config["vectorize_dim"] = vectorize_dim
    config["hidden_size"] = hidden_size
    config["num_layers"] = num_layers
    config["bidirectional"] = bidirectional

    return config


def create_casing_config(num_vocabs, vectorize_dim):
    config = {}
    config["num_vocabs"] = num_vocabs
    config["vectorize_dim"] = vectorize_dim

    return config


def create_crf_config(target_size):
    config = {}
    config["target_size"] = target_size

    return config


def create_lstm_config(input_size, hidden_size, p_dropout):
    config = {}
    config["input_size"] = input_size
    config["hidden_size"] = hidden_size
    config["p_dropout"] = p_dropout

    return config



import sys
from pathlib import Path



FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))

from defaults.train_defaults import BEST_FILE_NAME
from .logger import log_epoch_result, log_saving_result

import time
from datetime import datetime
import torch
from torch.optim.lr_scheduler import ReduceLROnPlateau

class CustomCallback(object):
    """
    A Callback for Learning Rate Scheduler,
    Early Stopping, Saving Model,
    Logging, ...
    """
    def __init__(
        self, model_name, epochs, optimizer, factor, decay_patience, mode, es_patience
    ):
        self.lr_scheduler = ReduceLROnPlateau(
            optimizer=optimizer, mode=mode, factor=factor, patience=decay_patience
        )
        self.model_name = model_name
        self.epochs = epochs
        self.es_patience = es_patience
        self.begin_time = time.time()
        self.best_f1 = -1
        self.bad_epochs = 0

    def step(self, epoch, epoch_loss, f1, learning_rate, model_state_dict):
        epoch_time = time.time() - self.begin_time
        f1 = round(f1 * 100, 2)
        log_epoch_result(epoch, self.epochs, epoch_loss, f1, learning_rate, epoch_time)

        self.lr_scheduler.step(f1)

        if f1 > self.best_f1:
            self.bad_epochs = 0
            self.best_f1 = f1
            file_name = str(datetime.now()) + " f1: " + str(f1) + ".w"
            best_file_name = BEST_FILE_NAME
            saving_path = ROOT / "models" / self.model_name / "exp" / file_name
            saving_best_path = (
                ROOT / "models" / self.model_name / "exp" / best_file_name
            )
            torch.save(model_state_dict, saving_path)
            torch.save(model_state_dict, saving_best_path)
            log_saving_result(saving_path, saving_best_path)
        else:
            self.bad_epochs += 1

        if self.bad_epochs > self.es_patience:
            return 1
        else:
            return 0

    def reset_state(self):
        self.begin_time = time.time()



import os
import sys
from pathlib import Path
FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))
    
from api.api_defaults import *
from modules.defaults.tag_dict import IDX_TO_TAG
from modules.model_configs.configs import get_configs
from modules.ner.model import NERModel
from modules.prepare.data import load_from_array
    
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from uvicorn import run
import argparse
import torch
from pydantic import BaseModel

def _initialize_model(model_mode, device_mode):

    """
    Create NERModel Instance with:
        model_mode: type of model.
        device_mode: cpu(0) or gpu - cuda(1).
    """
    device_name = "cpu"
    device = torch.device(device_name)

    if device_mode == 1:
        device_name = "cuda" if torch.cuda.is_available() else "cpu"
        device = torch.device(device_name)

    model_configs = get_configs(model_mode)

    model = NERModel(
        model_configs["lstm"],
        model_configs["char_level_embedding"],
        model_configs["casing_embedding"],
        model_configs["crf"],
        device,
    ).to(device)
    weights_path = ROOT / "modules" / "models" / model_configs["model_name"] / "best.w"
    model.load_state_dict(torch.load(weights_path, map_location=device))
    model.eval()

    return model


def _initialize_server(model):
    """
    Create a Server Instance with apis from NERModel.
    """
    app = FastAPI()

    @app.get("/")
    async def root():
        return {"message": "Welcome to the Food Vision API!"}

    class Item(BaseModel):
        sequences: list

    @app.post("/predict/")
    async def predict(item: Item):
        dataloader = load_from_array(item.sequences)
        with torch.no_grad():
            all_predicts = []
            for i, batch in enumerate(dataloader):
                best_paths = model.forward(batch)
                seq_lengths = batch["features"]["seq_lengths"]
                idxs_to_tags = lambda idxs: [IDX_TO_TAG[idx] for idx in idxs]
                for j in range(len(batch)):
                    sample_predicts = best_paths[j][: seq_lengths[j]]
                    all_predicts.append(idxs_to_tags(sample_predicts))
            return all_predicts

    return app


def start(model_mode = MODEL_MODE, device_mode = USE_GPU, port = PORT):
    """
    Create server and run on a port.
    """
    model = _initialize_model(model_mode, device_mode)
    app = _initialize_server(model)
    port = int(os.environ.get("PORT", port))
    run(app, port=port)


def parse_opt():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-mode", type=int, default=MODEL_MODE, help="mode of model"
    )
    parser.add_argument(
        "--device-mode", type=int, default=USE_GPU, help="0: cpu - 1: cuda"
    )
    parser.add_argument("--port", type=int, default=PORT, help="localhost: port")
    return parser.parse_args()


if __name__ == "__main__":
    opt = parse_opt()
    start(opt.model_mode, opt.device_mode, opt.port)



import sys
from pathlib import Path

FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))
    

from constants.model_constants import *
from defaults.model_defaults import *
from defaults.tag_dict import *

import torch
import torch.nn as nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
torch.manual_seed(SEED)

class NERModel(nn.Module):
    def __init__(
        self,
        lstm_config,
        char_level_embedding_config,
        casing_embedding_config,
        crf_config,
        device,
    ):
        super(NERModel, self).__init__()
        self.device = device

        self.tag_to_idx = TAG_TO_IDX
        self.use_char_level_embedding = char_level_embedding_config != None
        self.use_casing_embedding = casing_embedding_config != None
        self.use_crf = crf_config != None

        # Initialize Layers base on config dict oject
        if self.use_char_level_embedding:
            self.embedding_mode = char_level_embedding_config["mode"]
            self.char_vectorize = nn.Embedding(
                num_embeddings=char_level_embedding_config["num_vocabs"],
                embedding_dim=char_level_embedding_config["vectorize_dim"],
            )

            if self.embedding_mode == "CNN":
                cnn_embedding_config = char_level_embedding_config

                self.char_level_embeds = nn.Conv2d(
                    in_channels=cnn_embedding_config["in_channels"],
                    out_channels=cnn_embedding_config["out_channels"],
                    kernel_size=cnn_embedding_config["kernel_size"],
                    padding=cnn_embedding_config["padding_size"],
                )

            elif self.embedding_mode == "LSTM":
                lstm_embedding_config = char_level_embedding_config
                self.char_level_embeds = nn.LSTM(
                    input_size=lstm_embedding_config["vectorize_dim"],
                    hidden_size=lstm_embedding_config["hidden_size"],
                    num_layers=lstm_embedding_config["num_layers"],
                    bidirectional=lstm_embedding_config["bidirectional"],
                    batch_first=True,
                )
        if self.use_casing_embedding:
            self.casing_embeds = nn.Embedding(
                num_embeddings=casing_embedding_config["num_vocabs"],
                embedding_dim=casing_embedding_config["vectorize_dim"],
            )

        self.dropout = nn.Dropout(lstm_config["p_dropout"])
        self.lstm = nn.LSTM(
            input_size=lstm_config["input_size"],
            hidden_size=lstm_config["hidden_size"],
            bidirectional=True,
            batch_first=True,
        )

        target_size = len(self.tag_to_idx)
        self.hidden2tag = nn.Linear(lstm_config["hidden_size"] * 2, target_size)

        if self.use_crf:
            self.tag_to_idx[START] = START_IDX
            self.tag_to_idx[STOP] = STOP_IDX

            target_size = crf_config["target_size"]

            self.transitions = nn.Parameter(torch.randn(target_size, target_size))
            self.transitions.data[:, self.tag_to_idx[START]] = IMPOSSIBLE
            self.transitions.data[self.tag_to_idx[STOP]] = IMPOSSIBLE

            self.hidden2tag = nn.Linear(lstm_config["hidden_size"] * 2, target_size)

    def _get_lstm_features(self, batch):
        """
        Fit batch and return the outputs of lstm layers.
        """
        embeds = batch["global_vectors"].to(self.device)
        if self.use_char_level_embedding:

            char_vectorize_outputs = self.char_vectorize(
                batch["char_level_vectors"].to(torch.int64).to(self.device)
            )

            if self.embedding_mode == "CNN":
                char_embeds = self.char_level_embeds(
                    char_vectorize_outputs.view(
                        -1,
                        1,
                        char_vectorize_outputs.shape[2],
                        char_vectorize_outputs.shape[3],
                    )
                )

                char_embeds = char_embeds.view(
                    char_vectorize_outputs.shape[0],
                    char_vectorize_outputs.shape[1],
                    char_embeds.shape[2],
                    -1,
                )

                char_embeds = nn.functional.max_pool2d(
                    char_embeds, kernel_size=(char_vectorize_outputs.size(2), 1)
                ).squeeze(dim=2)

                embeds = torch.cat([embeds, char_embeds], dim=2)

            elif self.embedding_mode == "LSTM":
                out, (char_embeds, cn) = self.char_level_embeds(
                    char_vectorize_outputs.view(
                        -1,
                        char_vectorize_outputs.shape[2],
                        char_vectorize_outputs.shape[3],
                    )
                )
                char_embeds = char_embeds.view(
                    char_vectorize_outputs.shape[0], char_vectorize_outputs.shape[1], -1
                )
                embeds = torch.cat([embeds, char_embeds], dim=2)

        if self.use_casing_embedding:
            casing_embeds = self.casing_embeds(
                batch["casing_vectors"].to(torch.int64).to(self.device)
            )

            embeds = torch.cat([embeds, casing_embeds], dim=2)

        embeds = self.dropout(embeds)

        packed_embeds = pack_padded_sequence(
            embeds, batch["seq_lengths"], batch_first=True
        )
        lstm_out, _ = self.lstm(packed_embeds)
        lstm_out, _ = pad_packed_sequence(lstm_out, batch_first=True)
        lstm_out = self.dropout(lstm_out)
        lstm_feats = self.hidden2tag(lstm_out)

        return lstm_feats

    def fit_and_compute_loss(self, batch):
        """
        Fit batch to model then return loss.
        """
        batch_lstm_feats = self._get_lstm_features(batch["features"])

        if self.use_crf:
            loss = self.crf_loss(
                batch_lstm_feats, batch["tags"], batch["features"]["seq_lengths"]
            )
            return loss
        else:
            loss = self.cross_entropy_loss(
                batch_lstm_feats,
                batch["tags"],
                torch.Tensor(batch["features"]["seq_lengths"]).to(self.device),
            )
            return loss

    def forward(self, batch):
        # dont confuse this with _forward_alg above.
        # Get the emission scores from the BiLSTM
        batch_lstm_feats = self._get_lstm_features(batch["features"])
        if self.use_crf:
            max_sequence_len = batch["features"]["seq_lengths"][0]
            batch_size = len(batch["features"]["seq_lengths"])
            masks = torch.zeros(batch_size, max_sequence_len)
            for i, length in enumerate(batch["features"]["seq_lengths"]):
                masks[i, :length] = 1
            # Find the best path, given the features.
            score, tag_seq = self.__viterbi_decode(
                batch_lstm_feats,
                masks[:, : batch_lstm_feats.size(1)].float().to(self.device),
            )

            return tag_seq

        else:
            output = torch.argmax(batch_lstm_feats, dim=2).tolist()
            return output

    def cross_entropy_loss(self, batch_lstm_feats, tags, seq_lengths):
        """
        Calculate average cross entropy loss of batch by masking
        the padding words.
        """
        flatten_output = batch_lstm_feats.view(-1, batch_lstm_feats.shape[2])
        flatten_true_tags = tags.view(-1, 1).long().to(self.device)

        log_probs_flat = nn.functional.log_softmax(flatten_output, dim=1)

        losses_flat = -torch.gather(log_probs_flat, dim=1, index=flatten_true_tags)

        bs = batch_lstm_feats.shape[0]
        max_len = batch_lstm_feats.shape[1]
        losses = losses_flat.view(bs, max_len)

        seq_range = torch.arange(0, max_len).long().expand(bs, max_len).to(self.device)
        expanded_len = seq_lengths.view(bs, 1).expand(bs, max_len)
        mask = (expanded_len > seq_range).float()

        losses = losses * mask.float()
        loss = losses.sum() / seq_lengths.float().sum()
        return loss

    def log_sum_exp(self, x):
        """calculate log(sum(exp(x))) = max(x) + log(sum(exp(x - max(x))))"""
        max_score = x.max(-1)[0]
        return max_score + (x - max_score.unsqueeze(-1)).exp().sum(-1).log()

    def crf_loss(self, features, tags, seq_lengths):
        """negative log likelihood loss
        B: batch size, L: sequence length, D: dimension
        :param features: [B, L, D]
        :param ys: tags, [B, L]
        :param masks: masks for padding, [B, L]
        :return: loss
        """
        masks = torch.ones_like(tags).to(self.device)
        for i, length in enumerate(seq_lengths):
            masks[i, :length] = 1
        L = features.size(1)
        masks_ = masks[:, :L]

        forward_score = self.__forward_algorithm(features, masks_)
        gold_score = self.__score_sentence(
            features, tags[:, :L].long().to(self.device), masks_
        )
        loss = (forward_score - gold_score).mean()
        return loss

    def __score_sentence(self, features, tags, masks):
        """Gives the score of a provided tag sequence
        :param features: [B, L, C]
        :param tags: [B, L]
        :param masks: [B, L]
        :return: [B] score in the log space
        """
        B, L, C = features.shape

        # emission score
        emit_scores = features.gather(dim=2, index=tags.unsqueeze(-1)).squeeze(-1)

        # transition score
        start_tag = torch.full(
            (B, 1), self.tag_to_idx[START], dtype=torch.long, device=tags.device
        )
        tags = torch.cat([start_tag, tags], dim=1)  # [B, L+1]
        trans_scores = self.transitions[tags[:, 1:], tags[:, :-1]]

        # last transition score to STOP tag
        last_tag = tags.gather(dim=1, index=masks.sum(1).long().unsqueeze(1)).squeeze(
            1
        )  # [B]
        last_score = self.transitions[self.tag_to_idx[STOP], last_tag]

        score = ((trans_scores + emit_scores) * masks).sum(1) + last_score
        return score

    def __viterbi_decode(self, features, masks):
        """decode to tags using viterbi algorithm
        :param features: [B, L, C], batch of unary scores
        :param masks: [B, L] masks
        :return: (best_score, best_paths)
            best_score: [B]
            best_paths: [B, L]
        """
        B, L, C = features.shape

        bps = torch.zeros(
            B, L, C, dtype=torch.long, device=features.device
        )  # back pointers

        # Initialize the viterbi variables in log space
        max_score = torch.full((B, C), IMPOSSIBLE, device=features.device)  # [B, C]
        max_score[:, self.tag_to_idx[STOP]] = 0

        for t in range(L):
            mask_t = masks[:, t].unsqueeze(1)  # [B, 1]
            emit_score_t = features[:, t]  # [B, C]

            # [B, 1, C] + [C, C]
            acc_score_t = max_score.unsqueeze(1) + self.transitions  # [B, C, C]
            acc_score_t, bps[:, t, :] = acc_score_t.max(dim=-1)
            acc_score_t += emit_score_t
            max_score = acc_score_t * mask_t + max_score * (
                1 - mask_t
            )  # max_score or acc_score_t

        # Transition to STOP_TAG
        max_score += self.transitions[self.tag_to_idx[STOP]]
        best_score, best_tag = max_score.max(dim=-1)

        # Follow the back pointers to decode the best path.
        best_paths = []
        bps = bps.cpu().numpy()
        for b in range(B):
            best_tag_b = best_tag[b].item()
            seq_len = int(masks[b, :].sum().item())

            best_path = [best_tag_b]
            for bps_t in reversed(bps[b, :seq_len]):
                best_tag_b = bps_t[best_tag_b]
                best_path.append(best_tag_b)
            # drop the last tag and reverse the left
            best_paths.append(best_path[-2::-1])

        return best_score, best_paths

    def __forward_algorithm(self, features, masks):
        """calculate the partition function with forward algorithm.
        TRICK: log_sum_exp([x1, x2, x3, x4, ...]) = log_sum_exp([log_sum_exp([x1, x2]), log_sum_exp([x3, x4]), ...])
        :param features: features. [B, L, C]
        :param masks: [B, L] masks
        :return:    [B], score in the log space
        """
        B, L, C = features.shape

        scores = torch.full((B, C), IMPOSSIBLE, device=features.device)  # [B, C]
        scores[:, self.tag_to_idx[START]] = 0.0
        trans = self.transitions.unsqueeze(0)  # [1, C, C]

        # Iterate through the sentence
        for t in range(L):
            emit_score_t = features[:, t].unsqueeze(2)  # [B, C, 1]
            score_t = (
                scores.unsqueeze(1) + trans + emit_score_t
            )  # [B, 1, C] + [1, C, C] + [B, C, 1] => [B, C, C]
            score_t = self.log_sum_exp(score_t)  # [B, C
            mask_t = masks[:, t].unsqueeze(1)  # [B, 1]
            scores = score_t * mask_t + scores * (1 - mask_t)
        scores = self.log_sum_exp(scores + self.transitions[self.tag_to_idx[STOP]])
        return scores



from setuptools import find_namespace_packages, find_packages, setup

VERSION = "1.0.1"
DESCRIPTION = "An NER API trained on CONLL2003"
LONG_DESCRIPTION = "An NER API trained on CONLL2003"

# Setting up
setup(
    name="kmthachner",
    version=VERSION,
    author="Nguyen Kim Thach",
    author_email="<kmthach.ai@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_namespace_packages(),
    include_package_data=True,
    package_data={"": ["*.txt",'*.json','*.w']},
    install_requires=["torch", "seqeval", "datasets", "fastapi", "uvicorn"],
)



import sys
from pathlib import Path

FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))

import datasets
from constants.dataset_constants import DATASET_NAME


def load_data():
    # Load raw data from datasets library
    raw_datasets = datasets.load_dataset(DATASET_NAME)
    return raw_datasets



import sys
from pathlib import Path

FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))

from constants.dataset_constants import *
""" 
Logging functions in train, validation, test phase
"""
def log_configuration(phase, model_name, train_config, device_name):
    """
    Logging final settings for train, validation and test phase:
        device, learning rate, metrics,...
    """
    if phase == "train":
        learning_rate = train_config["learning_rate"]
        batch_size = train_config["batch_size"]
        epochs = train_config["epochs"]
        print(
            f"Train model [{model_name}] on CONLL-2003 train dataset({TRAIN_NUM_SAMPLES} samples) : device = {device_name}, learning rate = {learning_rate}, batch size = {batch_size}, total epochs = {epochs} "
        )

    elif phase == "validation":
        print(
            f"Validate model [{model_name}] on CONLL-2003 validation dataset({VALIDATION_NUM_SAMPLES} samples) : device = {device_name}, metrics = f1"
        )

    elif phase == "test":
        print(
            f"Test model [{model_name}] on CONLL-2003 test dataset({TEST_NUM_SAMPLES} samples) : device = {device_name}, metrics = f1, precision, recall"
        )


def log_epoch_result(epoch, total_eppochs, epoch_loss, f1, learning_rate, epoch_time):
    print(
        f"Epoch {epoch+1}/{total_eppochs}: train_loss: {epoch_loss:.2f} - validation_f1: {f1:.2f} - lr: {learning_rate} - time: {epoch_time:.0f}s"
    )


def log_saving_result(saving_path, saving_best_path):
    print(f"Successfully save model to {saving_path}")
    print(f"Successfully save model to {saving_best_path}")


def log_test_result(f1, precision, recall, report):
    print(f"Result: F1: {f1:.2f} - Precision: {precision:.2f} - Recall: {recall:.2f}")
    print(report)



import sys
from pathlib import Path
FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))

from defaults.featuring_defaults import *

import torch
import numpy as np

class Featuring(object):
    """
    Use for transfrom raw data to features
    """

    def __init__(self, use_char_level, use_casing):
        self.use_char_level = use_char_level
        self.use_casing = use_casing
        embeddings_dict = {}

        # Read pretrain data and save into embeddings_dict
        with open(ROOT / PRETRAIN_PATH, "r", encoding="utf-8") as f:
            for line in f:
                values = line.split()
                word = values[0]
                vector = torch.tensor(np.asarray(values[1:], "float32"))
                embeddings_dict[word] = vector
        self.embeddings_dict = embeddings_dict

    def tokens_to_glove_vectors(self, tokens):
        # Convert tokens to GloVe
        word_to_vec = lambda word: self.embeddings_dict.get(
            word.lower(), torch.zeros(GLOVE_DIM)
        )

        global_vectors = torch.stack([word_to_vec(token) for token in tokens])
        return global_vectors

    def tokens_to_char_level_vectors(self, tokens):
        all_characters = ALL_CHARACTERS
        char_dict = {c: idx + 1 for idx, c in enumerate(all_characters)}

        token_to_seq = lambda token: [char_dict[char] for char in token]

        char_level_vectors = [token_to_seq(token) for token in tokens]
        return char_level_vectors

    def tokens_to_casing_vectors(self, tokens):
        def token_to_casing(token):
            casing = 1

            numDigits = 0
            for char in token:
                if char.isdigit():
                    numDigits += 1

            digitFraction = numDigits / float(len(token))

            if token.isdigit():  # Is a digit
                casing = 2
            elif digitFraction > 0.5:
                casing = 3
            elif token.islower():  # All lower case
                casing = 4
            elif token.isupper():  # All upper case
                casing = 5
            elif token[0].isupper():  # is a title, initial char upper, then all lower
                casing = 6
            elif numDigits > 0:
                casing = 7

            return casing

        casing_vectors = [token_to_casing(token) for token in tokens]
        return casing_vectors

    def preprocess(self, tokens, tags):
        converted_sample = {"features": {}}

        if tags != None:
            converted_sample["tags"] = tags
        global_vectors = self.tokens_to_glove_vectors(tokens)
        converted_sample["features"]["global_vectors"] = global_vectors
        if self.use_char_level:

            char_level_vectors = self.tokens_to_char_level_vectors(tokens)
            converted_sample["features"]["char_level_vectors"] = char_level_vectors

        if self.use_casing:
            casing_vectors = self.tokens_to_casing_vectors(tokens)
            converted_sample["features"]["casing_vectors"] = casing_vectors
        return converted_sample
